#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include "buffer.h"
#include "error.h"
#include "strerr.h"
#include "direntry.h"
#include "auto_ageout.h"
#include "auto_rmcmd.h"
#include "wait.h"
#include "sgetopt.h"
#include "str.h"
#include "time.h"

#define FATAL "cleanq: fatal: "
#define WARNING "cleanq: warning: "

static char *cmd[16];
static int args;
static int flaglog = 0;

void die_usage(void)
{
  strerr_die1x(100,"cleanq: usage: cleanq [-l]");
}

void outs(const char *s)
{
  if (buffer_puts(buffer_1,s) == -1)
    strerr_die2sys(111,FATAL,"unable to write stdout: ");
}

void log_file(const char *h, const char *d, const char *m)
{
  if (!flaglog) return;
  outs(h);
  outs(d);
  outs(": ");
  outs(m);
  outs("\n");
  buffer_flush(buffer_1);
}

void warn_file(const char *d, const char *w)
{
  log_file(WARNING,d,w);
}

void direrror(void)
{
  strerr_warn2(FATAL,"unable to read directory: ",&strerr_sys);
}

void remove(const char *d)
{
  DIR  *dir  =  0;
  int rmstat = -1;
  pid_t pid  = -1;

  /* Directory must be chdirable */
  if (chdir(d) == -1 || chdir("..") == -1) {
    strerr_warn4(WARNING,"unable to chdir to ",d,": ",&strerr_sys);
    return;
  }
  /* OK, go ahead and try to delete it. */

  /* prepare the final argument */
  cmd[args] = (char *)d;

  /* run the delete command */
  switch(pid = vfork()) {
    case -1:
        strerr_warn2(WARNING,"unable to fork: ",&strerr_sys);
	return;
    case 0:
	execv(*cmd, cmd);
        strerr_warn2(WARNING,"unable to exec: ",&strerr_sys);
	return;
  }

  /* Catch the exit status */
  if (wait_pid(&rmstat, pid) == -1)
  {
    strerr_warn2(WARNING,"unable to get status: ",&strerr_sys);
  }
  else if (wait_exitcode(rmstat) != 0)
  {
    buffer_puts(buffer_2,WARNING);
    buffer_putsflush(buffer_2,"delete command returned nonzero status.\n");
  }
}

void tryclean(const char *d)
{
  struct stat st;
  time_t now;
 
  /* 1. Ignore "." and ".." */
  if(!str_diff(d,".") || !str_diff(d,"..")) return;

  /* 2. Try unlinking non-directories. These should never exist. */
  if (stat(d,&st) == -1) {
    strerr_warn4(WARNING,"unable to stat ",d,": ",&strerr_sys);
    return;
  }
  if ((st.st_mode & S_IFMT)!=S_IFDIR) {
    warn_file(d,"not a directory");
    unlink(d);  /* If it fails, oh well. */
    return;
  }

  /* 3. Ignore files not beginning with "@" */
  if (d[0] != '@') {
    warn_file(d, "name doesn't start with '@'");
    return;
  }

  /* 4. Delete if not sticky. */
  if (!(st.st_mode & 01000)) { /* not sticky */
    log_file("deleting: ",d,"not sticky");
    remove(d);
    return;
  } 

  /* 5. Skip if it's not old enough. */
  now = time(NULL);
  if ((now - st.st_ctime) < 3*MAX_AGE) return; /* not old */

  /* Delete it. Optionally, log the action. */
  log_file("deleting: ",d,"too old");
  remove(d);
}

int main(int argc, char **argv)
{
  DIR *dir;
  direntry *d;
  int opt;

  /* Use sgetopt although there's only one option--so far. */
  while ((opt = getopt(argc, (const char *const *)argv,"l")) != opteof)
    switch(opt)
    {
     case 'l': flaglog = 1; break;
     default: die_usage();
    }
  if (flaglog) {outs("cleanq starting\n");buffer_flush(buffer_1);}

  /* Build an "rm" command for repeated use. */
  for (args = 0; auto_rmcmd[args] != 0; args++)
  {
    cmd[args] = auto_rmcmd[args];
  }
  cmd[args--] = 0;

  /* Open the current directory */
  dir = opendir(".");
  if (!dir) {
    direrror();
    _exit(1);
  }

  /* Scan it */
  for (;;) {
    errno = 0;
    d = readdir(dir);
    if (!d) break;
    tryclean(d->d_name);
  }
  if (errno) { direrror(); closedir(dir); _exit(1); }
  closedir(dir);
  _exit(0);
}
