#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "buffer.h"
#include "exit.h"
#include "open.h"
#include "fd.h"
#include "wait.h"
#include "seek.h"
#include "sig.h"
#include "auto_ageout.h"
#include "exitcodes.h"
#include "mkfn.h"
#include "auto_fnlen.h"
#include "auto_scan_stdin.h"
#include "env.h"
#include "strerr.h"


int do_ripmime();
int do_scan();
int do_cleanq();

extern char *const auto_qmailq[];

#define FATAL "qscanq: fatal: "
int flaglog = 0;
void err(const char *s)
{
  if (flaglog && (buffer_putsflush(buffer_2,s) == -1))
    strerr_die2sys(QQ_XTEMP,FATAL,"unable to write stderr: ");
}

/* Signal handler just sets a flag. */
int alarm_flag;
pid_t pid;
void alarm_handler(int sig)
{
  if (pid) kill(pid, SIGKILL);
  else _exit(QQ_XTEMP);
}

/* The current directory is now "work", under a subdirectory of the
   virus scanning queue. A message is on stdin, and its envelope on
   file descriptor 2. Our environment is empty. Our job is to extract
   the MIME parts of the message, and run the virus scanner in cwd.
 */
int main (int argc, char *argv[])
{
  int fdout=-1, fdin=-1;
  int n=0;
  char buf[256];
  buffer msgbuf;
  char fn[FN_BYTES];
  pid = 0;

  /* Set an alarm handler */
  alarm_flag = 0;
  sig_alarmcatch(alarm_handler);
  alarm(MAX_AGE);

  /* Check whether we should be logging errors */
  if (env_get("DEBUG")) flaglog = 1;

  /* Set up stdin with seekable copy of message */
  for (;;) {
    /* [ fn := timestamp.ppid.n ] */
    mkfn(fn, n++);

    /* [ stdin := copied to a file named timestamp.ppid.n in cwd ] */
    if ((fdout = open_excl(fn)) < 0) {
      if (flaglog)
        strerr_die4sys(QQ_XTEMP,FATAL,"could not create file '",fn,"': ");
      do_cleanq(); _exit(QQ_XWRTE);
    }
    buffer_init(&msgbuf,buffer_unixwrite,fdout,buf,sizeof(buf));
    if (buffer_copy(&msgbuf,buffer_0) < 0) {
      if (flaglog)
        strerr_die2sys(QQ_XTEMP,FATAL,"could not copy stdin: ");
      do_cleanq();
      _exit(QQ_XWRTE);
    }
    if (buffer_flush(&msgbuf)<0) {
      if (flaglog)
        strerr_die2sys(QQ_XTEMP,FATAL,"could not flush stdin: ");
      do_cleanq();
      _exit(QQ_XWRTE);
    }
    close(fdout);
  
    /* On the second pass, leave the file there. */
    if (n == 2) break;

    /* [ stdin := descriptor pointing to msg file, opened for reading ] */
    if ((fdin = open_read(fn))<0) {
      if (flaglog)
        strerr_die4sys(QQ_XTEMP,FATAL,"could not open '",fn,"' for reading: ");
      do_cleanq();
      _exit(QQ_XREAD);
    }
    if (fd_move(0, fdin)<0) {
      if (flaglog)
        strerr_die2sys(QQ_XTEMP,FATAL,"fdmove failed: ");
      do_cleanq();
      _exit(QQ_XREAD);
    }
    unlink(fn);

    /* If we scan stdin, we're ready now. */
    if (SCAN_STDIN) break;
  }
  seek_begin(0);

  /* Run ripmime now, to extract the message parts.
   * [ stdin := extracted to files under cwd ]
   */
  if (!do_ripmime()) {
    err("MIME extraction failed\n");
    do_cleanq();
    _exit(QQ_XREAD);
  } 
  seek_begin(0);

  /* Run antivir now, to check the current directory for viruses.
   */
  switch(do_scan()) {
    case(EX_ALLOK):
      break;
    case(EX_VIRUS): /* Found a virus */
      do_cleanq(); _exit(QQ_XPERM); /* drop the message */
    default:  /* some problem w/scan */
      err("some problem occurred while scanning the message\n");
      do_cleanq(); _exit(QQ_XTEMP);
  } 

  /* all done! */
  do_cleanq();
  seek_begin(0);	/* Reset stdin for next in chain */
  if (argc>1) execv(argv[1], argv+1);
  else execv(*auto_qmailq, auto_qmailq);
  _exit(QQ_XTEMP);
}
